"""
This module contains the API for the game.
"""

import pygame

from .generators import *
from .events import *
from .utils import *
from .random import *


pygame.init()  # pylint: disable=no-member
